// Copyright (c) 2005..2010 Dirk Jagdmann <doj@cubic.org>

// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.

// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:

//     1. The origin of this software must not be misrepresented; you
//        must not claim that you wrote the original software. If you use
//        this software in a product, an acknowledgment in the product
//        documentation would be appreciated but is not required.

//     2. Altered source versions must be plainly marked as such, and
//        must not be misrepresented as being the original software.

//     3. This notice may not be removed or altered from any source
//        distribution.

#include <wav.h>
#include <cctype>
#include <cstdio>
#include <string>
#include <vector>
#include <map>
#include <cassert>
#include <errno.h>
#include <cstring>
#include <sizeof_array.h>
#include <stdexcept>

#include "type.cpp"

namespace
{
  unsigned bytes_per_second;

  class RIFF;

#define READp(var) if(fread(var, sizeof(var), 1, f) != 1) throw std::runtime_error(CURRENT_CLASS ": could not read " #var " tag")
#define READv(var) if(fread(& var, sizeof(var), 1, f) != 1) throw std::runtime_error(CURRENT_CLASS ": could not read " #var " tag")

#undef  CURRENT_CLASS
#define CURRENT_CLASS "Chunk"
  class Chunk
  {
  protected:
    FILE *f;
    unsigned indent;

  public:
    char name[4];
    uint32_t length;

    virtual ~Chunk() { }

    Chunk(const Chunk& c) :
      f(c.f),
      indent(c.indent),
      length(c.length)
    {
      memcpy(name, c.name, sizeof(name));
    }

    Chunk(FILE *f_, unsigned indent_) :
      f(f_),
      indent(indent_)
    {
      READp(name);
      READv(length);
      // length must be even
      if(length&1) ++length;
    }

    void printIndent()
    {
      if(indent>1024) indent=1024;
      for(unsigned i=0; i!=indent; ++i) putchar(' ');
    }

    virtual void print()
    {
      printIndent();
      printf("%c%c%c%c %u bytes", name[0],name[1],name[2],name[3], length);
    }

    virtual void advance() { }

    bool operator == (const std::string& s)
    {
      if(s.size() != 4) return false;
      return strncmp(s.c_str(), name, sizeof(name)) == 0;
    }

    static Chunk* factory(FILE *f, unsigned indent=0);
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "RIFF"
  class RIFF : public Chunk
  {
  public:
    char type[4];

    RIFF(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");

      READp(type);
    }

    virtual void print()
    {
      Chunk::print();
      printf(" type:%c%c%c%c", type[0],type[1],type[2],type[3]);
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "fmt "
  class fmt : public Chunk
  {
  public:
    uint16_t type;		/* WAVE_FORMAT_... */
    uint16_t channels;		/* commonly 1,2 but can be >2 */
    uint32_t rate;
    uint32_t bytesPerSecond;	/* rate*bytesPerFrame */
    union {
      uint16_t blockAlign;
      uint16_t bytesPerFrame;	/* channels*((bitsPerSample+7)/8) */
    };
    uint16_t bitsPerSample;	/* 8,12,16,24,32 */

    fmt(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS "constructed with wrong chunk");

      indent+=2;

      READv(type);
      READv(channels);
      READv(rate);
      READv(bytesPerSecond);
      READv(bytesPerFrame);
      READv(bitsPerSample);
    }

    virtual void print()
    {
      Chunk::print();
      printf(" type:\"%s\" channels:%i rate:%i bytesPerSecond:%i bytesPerFrame:%i bitsPerSample:%i",
	     type2string(type), channels, rate, bytesPerSecond, bytesPerFrame, bitsPerSample);
      bytes_per_second=bytesPerSecond;
    }

    virtual void advance()
    {
      int a=length-16;
      if(a>0)
	fseek(f, a, SEEK_CUR);
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "data"
  class data : public Chunk
  {
    bool oversized;

  public:
    data(const Chunk* c) :
      Chunk(*c),
      oversized(false)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;

      // check for oversized wav
      if(length>=2147483647)
	{
	  oversized=true;
	  fseek(f, 0, SEEK_END);
	}
    }

    virtual void print()
    {
      Chunk::print();
      if(bytes_per_second)
	{
	  int sec=length/bytes_per_second;
	  const int s=sec%60; sec/=60;
	  const int m=sec%60; sec/=60;
	  printf(" = %i:%02i:%02i", sec,m,s);
	  if(oversized)
	    printf(" maximum filesize reached!");
	}
    }

    virtual void advance()
    {
      fseek(f, length, SEEK_CUR);
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "LIST"
  class LIST : public Chunk
  {
  public:
    char id[4];
    long fileoffset;

    LIST(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;

      fileoffset=ftell(f)+length;

      READp(id);
    }
    virtual void print()
    {
      Chunk::print();
      printf(" %c%c%c%c ", id[0], id[1], id[2], id[3]);

    }
    virtual void advance()
    {
      while(!feof(f) && ftell(f) < fileoffset)
	{
	  Chunk *c=Chunk::factory(f, indent+2);
	  if(c)
	    {
	      c->print();
	      putchar('\n');
	      c->advance();
	    }
	}
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "PEAK"
  // http://music.calarts.edu/%7Etre/PeakChunk.html
  class PEAK : public Chunk
  {
  public:
    uint32_t version;
    uint32_t timeStamp;

#pragma pack(push, 1)
    struct PositionPeak
    {
      float val;
      uint32_t pos;
    };
#pragma pack(pop)

    PositionPeak *peak;
    unsigned peakNum;

    char *data;
    int datalength;

    virtual ~PEAK() { delete [] data; delete [] peak; }

    PEAK(const Chunk* c) :
      Chunk(*c),
      peak(0),
      peakNum(0),
      data(0),
      datalength(0)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;

      READv(version);
      if(version == 1)
	{
	  READv(timeStamp);
	  peakNum=(length-sizeof(version)-sizeof(timeStamp))/sizeof(PositionPeak);
	  if(peakNum>0)
	    {
	      peak=new PositionPeak[peakNum];
	      if(fread(peak, sizeof(PositionPeak), peakNum, f) != peakNum)
		throw std::runtime_error("PEAK: could not read peaks");
	    }
	}
      else
	{
	  datalength=length-sizeof(version);
	  if(datalength > 0)
	    {
	      data=new char[datalength];
	      if(fread(data, datalength, 1, f) != 1)
		throw std::runtime_error("PEAK: could not read data");
	    }
	}
    }
    virtual void print()
    {
      Chunk::print();
      printf(" version:%i", version);
      if(version == 1)
	{
	  time_t t=timeStamp;
	  char *ts=ctime(&t);
	  // replace newline with "
	  char *p=strchr(ts, '\n');
	  if(p) { *p++='"'; *p=0; }
	  printf(" \"%s", ts);

	  for(unsigned i=0; i!=peakNum; ++i)
	    printf(" %i:%.02f ", peak[i].pos, peak[i].val);
	}
      else
	{
	  printf(" data: ");
	  for(int i=0; i<datalength; ++i)
	    putchar(std::isprint(data[i]) ? data[i] : '.');
	}
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "DISP"
  class DISP : public Chunk
  {
  public:
    char *data;
    int datalength;
    uint32_t type;

    virtual ~DISP() { delete [] data; }

    DISP(const Chunk* c) :
      Chunk(*c),
      data(0)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;

      READv(type);

      datalength=length-sizeof(type);
      if(datalength>0)
	{
	  data=new char[datalength];
	  if(fread(data, datalength, 1, f) != 1)
	    throw std::runtime_error("DISP: could not read data");
	}
    }
    virtual void print()
    {
      Chunk::print();
      switch(type)
	{
	case 1: printf(" text: %s", data); break;
	case 2: printf(" bitmap: "); goto print_generic;
	default: printf(" unknown data: ");
	print_generic:
	  for(int i=0; i<datalength; ++i)
	    putchar(std::isprint(data[i]) ? data[i] : '.');
	}
    }
  };

  class cue : public Chunk
  {
  public:

    struct cuePoint
    {
      uint32_t identifier;
      uint32_t position;
      char     id[4];
      uint32_t chunkStart;
      uint32_t blockStart;
      uint32_t sampleOffset;

      void read(FILE *f)
      {
#undef  CURRENT_CLASS
#define CURRENT_CLASS "cuePoint"
	READv(identifier);
	READv(position);
	READp(id);
	READv(chunkStart);
	READv(blockStart);
	READv(sampleOffset);
      }

      void print(int indent)
      {
	for(int i=0; i!=indent; ++i)
	  putchar(' ');
	printf("cuePoint identifier:%i position:%i id:%c%c%c%c chunkStart:%i blockStart:%i sampleOffset:%i",
	       identifier, position, id[0],id[1],id[2],id[3], chunkStart, blockStart, sampleOffset);
      }
    };

    typedef std::vector<cuePoint> v_t;

    uint32_t num;
    v_t points;

#undef  CURRENT_CLASS
#define CURRENT_CLASS "cue "
    cue(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS "constructed with wrong chunk");
      indent+=2;

      READv(num);
      assert(num < 100000);
      if(num)
	{
	  points.resize(num);
	  for(unsigned i=0; i!=num; ++i)
	    points[i].read(f);
	}
      else
	{
	  int s=length-4;
	  if(s>0)
	    fseek(f, s, SEEK_CUR);
	}
    }

    virtual void print()
    {
      Chunk::print();
      printf(" num:%i", (int)num);
      if(num>0)
	printf("\n");
      for(unsigned i=0; i!=num; ++i)
	points[i].print(indent+2);
    }
  };

  class plst : public Chunk
  {
  public:

    struct Segment
    {
      uint32_t identifier;
      uint32_t length;
      uint32_t repeats;

      void read(FILE *f)
      {
#undef  CURRENT_CLASS
#define CURRENT_CLASS "Segment"
	READv(identifier);
	READv(length);
	READv(repeats);
      }

      void print(int indent)
      {
	for(int i=0; i!=indent; ++i)
	  putchar(' ');
	printf("Segment identifier:%i length:%i repeats:%i\n",
	       identifier, length, repeats);
      }
    };

    typedef std::vector<Segment> v_t;

    uint32_t num;
    v_t segments;

#undef  CURRENT_CLASS
#define CURRENT_CLASS "plst"
    plst(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS" constructed with wrong chunk");
      indent+=2;

      READv(num);
      assert(num < 100000);
      if(num)
	{
	  segments.resize(num);
	  for(unsigned i=0; i!=num; ++i)
	    segments[i].read(f);
	}
      else
	{
	  int s=length-4;
	  if(s>0)
	    fseek(f, s, SEEK_CUR);
	}
    }

    virtual void print()
    {
      Chunk::print();
      printf(" num:%i", (int)num);
      for(unsigned i=0; i!=num; ++i)
	segments[i].print(indent+2);
    }
  };

  class TextChunk : public Chunk
  {
    std::map<std::string, std::string> idmap;
    void init();

  public:
    char *data;
    char id[5];

    virtual ~TextChunk() { delete [] data; }

    TextChunk(const Chunk* c) :
      Chunk(*c)
    {
      init();

      for(int i=0; i<4; ++i)
	id[i]=std::tolower(name[i]);
      id[4]=0;

      if(idmap.count(id) == 0)
	throw std::runtime_error("TextChunk constructed with wrong chunk");

      indent+=2;

      data=new char[length];
      if(fread(data, length, 1, f) != 1)
	throw std::runtime_error("TextChunk: could not read data");
    }
    virtual void print()
    {
      Chunk::print();
      printf(": %s: %s", idmap[id].c_str(), data);
    }
  };

  void TextChunk::init()
  {
    if(idmap.begin() != idmap.end()) return;
    idmap["iarl"]="Archival location";
    idmap["iart"]="Artist";
    idmap["icms"]="Commissioned";
    idmap["icmt"]="Comments";
    idmap["icop"]="Copyright";
    idmap["icrd"]="Creation date";
    idmap["ieng"]="Engineer";
    idmap["ignr"]="Genre";
    idmap["ikey"]="Keywords";
    idmap["inam"]="Name";
    idmap["imed"]="Medium";
    idmap["iprd"]="Product";
    idmap["isbj"]="Subject";
    idmap["isft"]="Software";
    idmap["isrc"]="Source";
    idmap["isrf"]="Original medium";
    idmap["itch"]="Technician";
  }

#define READs(var,size) do { \
  if(fread(& var, size, 1, f) != 1) throw std::runtime_error(CURRENT_CLASS ": could not read " #var " tag"); \
  var[size]=0; \
} while(0)

#undef  CURRENT_CLASS
#define CURRENT_CLASS "cart"
  class cart : public Chunk
  {
  public:
    struct cart_timer_tag
    {
      char dwUsage[4];
      int32_t dwValue;
    };

    char Version[4+1];
    char Title[64+1];
    char Artist[64+1];
    char CutID[64+1];
    char ClientID[64+1];
    char Category[64+1];
    char Classification[64+1];
    char OutCue[64+1];
    char StartDate[10+1];
    char StartTime[8+1];
    char EndDate[10+1];
    char EndTime[8+1];
    char ProducerAppID[64+1];
    char ProducerAppVersion[64+1];
    char UserDef[64+1];
    int32_t dwLevelReference;
    cart_timer_tag PostTimer[8];
    char Reserved[276+1];
    char URL[1024+1];
    char *TagText;

    cart(const Chunk* c) :
      Chunk(*c),
      TagText(0)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS" constructed with wrong chunk");
      indent+=2;

      READs(Version, 4);
      READs(Title, 64);
      READs(Artist, 64);
      READs(CutID, 64);
      READs(ClientID, 64);
      READs(Category, 64);
      READs(Classification, 64);
      READs(OutCue, 64);
      READs(StartDate, 10);
      READs(StartTime, 8);
      READs(EndDate, 10);
      READs(EndTime, 8);
      READs(ProducerAppID, 64);
      READs(ProducerAppVersion, 64);
      READs(UserDef, 64);
      READv(dwLevelReference);

      for(unsigned i=0; i<sizeof_array(PostTimer); ++i)
	{
#undef  CURRENT_CLASS
#define CURRENT_CLASS "cart_timer_tag"
	  READp(PostTimer[i].dwUsage);
	  READv(PostTimer[i].dwValue);
	}
#undef  CURRENT_CLASS
#define CURRENT_CLASS "cart"
      READs(Reserved, 276);
      READs(URL, 1024);

      int TagTextLen=length - 4-64-64-64-64-64-64-64-10-8-10-8-64-64-64-4-(8*(4+4))-276-1024;
      if(TagTextLen > 0)
	{
	  TagText = new char[TagTextLen+1];
	  if(fread(TagText, TagTextLen, 1, f) != 1)
	    throw std::runtime_error(CURRENT_CLASS ": could not read TagText tag");
	  TagText[TagTextLen]=0;
	}
    }

    virtual void print()
    {
      Chunk::print(); putchar('\n');
      indent+=2;
      printIndent(); printf("Version:          %s\n", Version);
      printIndent(); printf("Title:            %s\n", Title);
      printIndent(); printf("Artist:           %s\n", Artist);
      printIndent(); printf("CutID:            %s\n", CutID);
      printIndent(); printf("ClientID:         %s\n", ClientID);
      printIndent(); printf("Category:         %s\n", Category);
      printIndent(); printf("Classification:   %s\n", Classification);
      printIndent(); printf("OutCue:           %s\n", OutCue);
      printIndent(); printf("Start-End:        %s %s - %s %s\n", StartDate, StartTime, EndDate, EndTime);
      printIndent(); printf("ProducerAppID:    %s\n", ProducerAppID);
      printIndent(); printf("ProducerAppVer:   %s\n", ProducerAppVersion);
      printIndent(); printf("UserDef:          %s\n", UserDef);
      printIndent(); printf("dwLevelReference: %i\n", dwLevelReference);
      for(unsigned i=0; i<sizeof_array(PostTimer); ++i)
	{
	  if(!strncmp(PostTimer[i].dwUsage, "\0\0\0\0", 4)) continue;
	  printIndent();
	  printf("TimerTag %i:       %c%c%c%c %i\n", i,
		 isprint(PostTimer[i].dwUsage[0]) ? PostTimer[i].dwUsage[0] : '.',
		 isprint(PostTimer[i].dwUsage[1]) ? PostTimer[i].dwUsage[1] : '.',
		 isprint(PostTimer[i].dwUsage[2]) ? PostTimer[i].dwUsage[2] : '.',
		 isprint(PostTimer[i].dwUsage[3]) ? PostTimer[i].dwUsage[3] : '.',
		 PostTimer[i].dwValue
		 );
	}
      printIndent(); printf("URL:              %s", URL);
      if(TagText)
	{
	  printf("\n");
	  printIndent(); printf("TagText:          %s", TagText);
	}
    }
  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "bext"
  class bext : public Chunk
  {
  public:
    char Description[256+1];
    char Originator[32+1];
    char OriginatorReference[32+1];
    char OriginationDate[10+1];
    char OriginationTime[8+1];
    int32_t TimeReferenceLow;
    int32_t TimeReferenceHigh;
    int16_t Version;
    int8_t UMID[64];
    char Reserved[190];
    char *CodingHistory;

    bext(const Chunk* c) :
      Chunk(*c),
      CodingHistory(0)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;

      READs(Description, 256);
      READs(Originator, 32);
      READs(OriginatorReference, 32);
      READs(OriginationDate, 10);
      READs(OriginationTime, 8);
      READv(TimeReferenceLow);
      READv(TimeReferenceHigh);
      READv(Version);
      READp(UMID);
      READp(Reserved);

      int CodingHistoryLen=length - 256-32-32-10-8-4-4-2-64-190;
      if(CodingHistoryLen > 0)
	{
	  CodingHistory = new char[CodingHistoryLen+1];
	  if(fread(CodingHistory, CodingHistoryLen, 1, f) != 1)
	    throw std::runtime_error(CURRENT_CLASS ": could not read CodingHistory tag");
	  CodingHistory[CodingHistoryLen]=0;
	}
    }

    virtual void print()
    {
      Chunk::print(); putchar('\n');
      indent+=2;
      printIndent(); printf("Description:    %s\n", Description);
      printIndent(); printf("Originator:     %s\n", Originator);
      printIndent(); printf("OriginatorRef:  %s\n", OriginatorReference);
      printIndent(); printf("Date Time:      %s %s\n", OriginationDate, OriginationTime);
      printIndent(); printf("TimeRef Lo/Hi:  %i %i\n", TimeReferenceLow, TimeReferenceHigh);
      printIndent(); printf("Version:        %i", Version);
      if(CodingHistory)
	{
	  printf("\n");
	  printIndent(); printf("Coding History: %s", CodingHistory);
	}
    }

  };

#undef  CURRENT_CLASS
#define CURRENT_CLASS "fact"
  class fact : public Chunk
  {
  public:
    int32_t sampleLength;

    fact(const Chunk* c) :
      Chunk(*c)
    {
      if(strncmp(name, CURRENT_CLASS, 4))
	throw std::runtime_error(CURRENT_CLASS " constructed with wrong chunk");
      indent+=2;
      READv(sampleLength);
    }

    virtual void print()
    {
      Chunk::print();
      printf(" sampleLength: %i", sampleLength);
    }

  };

#undef READp
#undef READv
#undef READs
#undef CURRENT_CLASS

  Chunk* Chunk::factory(FILE *f, unsigned indent)
  {
    try
      {
	Chunk *ret=0;
	Chunk *c=new Chunk(f, indent);
	if(*c == "RIFF")
	  ret=new RIFF(c);
	else if(*c == "cue ")
	  ret=new cue(c);
	else if(*c == "data")
	  ret=new data(c);
	else if(*c == "fmt ")
	  ret=new fmt(c);
	else if(*c == "plst")
	  ret=new plst(c);
	else if(*c == "DISP")
	  ret=new DISP(c);
	else if(*c == "LIST")
	  ret=new LIST(c);
	else if(*c == "PEAK")
	  ret=new PEAK(c);
	else if(*c == "cart")
	  ret=new cart(c);
	else if(*c == "bext")
	  ret=new bext(c);
	else if(*c == "fact")
	  ret=new fact(c);
	else
	  {
	    try {
	      return new TextChunk(c);
	    }
	    catch (std::runtime_error&)
	      {
		c->print();
		printf(" (unknown tag)");
		char buf[256];
		if(c->length <= sizeof(buf))
		  {
		    printf(": ");
		    if(fread(buf, c->length, 1, f) == 1)
		      for(unsigned i=0; i<c->length; ++i)
			putchar(std::isprint(buf[i]) ? buf[i] : '.');
		  }
		else
		  fseek(f, c->length, SEEK_CUR);
		putchar('\n');
	      }
	  }

	delete c;
	return ret;
      }
    catch (std::runtime_error&)
      {
	return 0;
      }
    return 0;
  }

} // namespace

int main(int argc, char **argv)
{
  if(argc < 2)
    {
      printf("usage: wavinfo filename\n");
      return 1;
    }

  for(int i=1; i!=argc; ++i)
    {
      FILE *f=fopen(argv[i], "rb");
      if(f)
	{
	  puts(argv[i]);
	  while(!feof(f))
	    {
	      Chunk *c=Chunk::factory(f);
	      if(c)
		{
		  c->print();
		  putchar('\n');
		  c->advance();
		}
	    }
	  fclose(f);
	  putchar('\n');
	}
      else
	printf("could not open %s:%s\n", argv[i], strerror(errno));
    }
  return 0;
}
